@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 2005-2006 Douglas Gilbert.
+ * Copyright (c) 2005-2008 Douglas Gilbert.
  * All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
@@ -32,26 +32,31 @@
 #include <string.h>
 #include <errno.h>
 #include <ctype.h>
+#define __STDC_FORMAT_MACROS 1
+#include <inttypes.h>
 
 #include "sdparm.h"
 #include "sg_lib.h"
 #include "sg_cmds_basic.h"
 
 /* sdparm_vpd.c : does mainly VPD page processing associated with the
- * INQUIRY SCSI command.
+ * INQUIRY SCSI command. Roughly in sync with spc4r15.
  */
 
 /* Prints outs an abridged set of device identification designators
    selected by association, designator type and/or code set. */
-static int decode_dev_ids_quiet(unsigned char * buff, int len,
-                                int m_assoc, int m_desig_type,
-                                int m_code_set)
+static int
+decode_dev_ids_quiet(unsigned char * buff, int len, int m_assoc,
+                     int m_desig_type, int m_code_set)
 {
-    int m, p_id, c_set, piv, assoc, desig_type, i_len;
-    int c_id, d_id, naa, vsi, off, u;
+    int m, p_id, c_set, piv, assoc, desig_type, i_len, is_sas;
+    int naa, off, u, rtp;
     const unsigned char * ucp;
     const unsigned char * ip;
+    unsigned char sas_tport_addr[8];
 
+    rtp = 0;
+    memset(sas_tport_addr, 0, sizeof(sas_tport_addr));
     off = -1;
     while ((u = sg_vpd_dev_id_iter(buff, len, &off, m_assoc, m_desig_type,
                                    m_code_set)) == 0) {
@@ -66,6 +71,7 @@ static int decode_dev_ids_quiet(unsigned char * buff, int len,
         p_id = ((ucp[0] >> 4) & 0xf);
         c_set = (ucp[0] & 0xf);
         piv = ((ucp[1] & 0x80) ? 1 : 0);
+        is_sas = (piv && (6 == p_id)) ? 1 : 0;
         assoc = ((ucp[1] >> 4) & 0x3);
         desig_type = (ucp[1] & 0xf);
         switch (desig_type) {
@@ -76,7 +82,7 @@ static int decode_dev_ids_quiet(unsigned char * buff, int len,
         case 2: /* EUI-64 based */
             if ((8 != i_len) && (12 != i_len) && (16 != i_len))
                 fprintf(stderr, "      << expect 8, 12 and 16 byte "
-                        "ids, got %d>>\n", i_len);
+                        "EUI, got %d>>\n", i_len);
             printf("0x");
             for (m = 0; m < i_len; ++m)
                 printf("%02x", (unsigned int)ip[m]);
@@ -84,33 +90,20 @@ static int decode_dev_ids_quiet(unsigned char * buff, int len,
             break;
         case 3: /* NAA */
             if (1 != c_set) {
-                fprintf(stderr, "      << expected binary code_set (1)>>\n");
+                fprintf(stderr, "      << unexpected code set %d for "
+                        "NAA>>\n", c_set);
                 dStrHex((const char *)ip, i_len, 0);
                 break;
             }
             naa = (ip[0] >> 4) & 0xff;
-            if (! ((2 == naa) || (5 == naa) || (6 == naa))) {
-                fprintf(stderr, "      << expected naa [0x%x]>>\n", naa);
+            if ((naa < 2) || (naa > 6) || (4 == naa)) {
+                fprintf(stderr, "      << unexpected NAA [0x%x]>>\n", naa);
                 dStrHex((const char *)ip, i_len, 0);
                 break;
             }
-            if (2 == naa) {
+            if (2 == naa) {             /* NAA IEEE Extended */
                 if (8 != i_len) {
-                    fprintf(stderr, "      << expected NAA 2 identifier "
-                            "length: 0x%x>>\n", i_len);
-                    dStrHex((const char *)ip, i_len, 0);
-                    break;
-                }
-                d_id = (((ip[0] & 0xf) << 8) | ip[1]);
-                c_id = ((ip[2] << 16) | (ip[3] << 8) | ip[4]);
-                vsi = ((ip[5] << 16) | (ip[6] << 8) | ip[7]);
-                printf("0x");
-                for (m = 0; m < 8; ++m)
-                    printf("%02x", (unsigned int)ip[m]);
-                printf("\n");
-            } else if (5 == naa) {
-                if (8 != i_len) {
-                    fprintf(stderr, "      << expected NAA 5 identifier "
+                    fprintf(stderr, "      << unexpected NAA 2 identifier "
                             "length: 0x%x>>\n", i_len);
                     dStrHex((const char *)ip, i_len, 0);
                     break;
@@ -119,9 +112,37 @@ static int decode_dev_ids_quiet(unsigned char * buff, int len,
                 for (m = 0; m < 8; ++m)
                     printf("%02x", (unsigned int)ip[m]);
                 printf("\n");
-            } else if (6 == naa) {
+            } else if ((3 == naa ) || (5 == naa)) {
+                /* NAA=3 Locally assigned; NAA=5 IEEE Registered */
+                if (8 != i_len) {
+                    fprintf(stderr, "      << unexpected NAA 3 or 5 "
+                            "identifier length: 0x%x>>\n", i_len);
+                    dStrHex((const char *)ip, i_len, 0);
+                    break;
+                }
+                if ((0 == is_sas) || (1 != assoc)) {
+                    printf("0x");
+                    for (m = 0; m < 8; ++m)
+                        printf("%02x", (unsigned int)ip[m]);
+                    printf("\n");
+                } else if (rtp) {
+                    printf("0x");
+                    for (m = 0; m < 8; ++m)
+                        printf("%02x", (unsigned int)ip[m]);
+                    printf(",0x%x\n", rtp);
+                    rtp = 0;
+                } else {
+                    if (sas_tport_addr[0]) {
+                        printf("0x");
+                        for (m = 0; m < 8; ++m)
+                            printf("%02x", (unsigned int)sas_tport_addr[m]);
+                        printf("\n");
+                    }
+                    memcpy(sas_tport_addr, ip, sizeof(sas_tport_addr));
+                }
+            } else if (6 == naa) {      /* NAA IEEE Registered extended */
                 if (16 != i_len) {
-                    fprintf(stderr, "      << expected NAA 6 identifier "
+                    fprintf(stderr, "      << unexpected NAA 6 identifier "
                             "length: 0x%x>>\n", i_len);
                     dStrHex((const char *)ip, i_len, 0);
                     break;
@@ -133,8 +154,19 @@ static int decode_dev_ids_quiet(unsigned char * buff, int len,
             }
             break;
         case 4: /* Relative target port */
+            if ((0 == is_sas) || (1 != c_set) || (1 != assoc) || (4 != i_len))
+                break;
+            rtp = ((ip[2] << 8) | ip[3]);
+            if (sas_tport_addr[0]) {
+                printf("0x");
+                for (m = 0; m < 8; ++m)
+                    printf("%02x", (unsigned int)sas_tport_addr[m]);
+                printf(",0x%x\n", rtp);
+                memset(sas_tport_addr, 0, sizeof(sas_tport_addr));
+                rtp = 0;
+            }
             break;
-        case 5: /* Target port group */
+        case 5: /* (primary) Target port group */
             break;
         case 6: /* Logical unit group */
             break;
@@ -156,8 +188,14 @@ static int decode_dev_ids_quiet(unsigned char * buff, int len,
             break;
         }
     }
+    if (sas_tport_addr[0]) {
+        printf("0x");
+        for (m = 0; m < 8; ++m)
+            printf("%02x", (unsigned int)sas_tport_addr[m]);
+        printf("\n");
+    }
     if (-2 == u) {
-        fprintf(stderr, "VPD page error: short designator around "
+        fprintf(stderr, "VPD page error: short designator near "
                 "offset %d\n", off);
         return SG_LIB_CAT_MALFORMED;
     }
@@ -166,16 +204,18 @@ static int decode_dev_ids_quiet(unsigned char * buff, int len,
 
 /* Prints outs device identification designators selected by association,
    designator type and/or code set. */
-static int decode_dev_ids(const char * print_if_found, unsigned char * buff,
-                          int len, int m_assoc, int m_desig_type,
-                          int m_code_set, int long_out, int quiet)
+static int
+decode_dev_ids(const char * print_if_found, unsigned char * buff, int len,
+               int m_assoc, int m_desig_type, int m_code_set, int long_out,
+               int quiet)
 {
     int m, p_id, c_set, piv, assoc, desig_type, i_len;
     int ci_off, c_id, d_id, naa, vsi, printed, off, u;
-    unsigned long long vsei;
-    unsigned long long id_ext;
+    uint64_t vsei;
+    uint64_t id_ext;
     const unsigned char * ucp;
     const unsigned char * ip;
+    char b[64];
 
     if (quiet)
         return decode_dev_ids_quiet(buff, len, m_assoc, m_desig_type,
@@ -203,10 +243,11 @@ static int decode_dev_ids(const char * print_if_found, unsigned char * buff,
         }
         if (NULL == print_if_found)
             printf("  %s:\n", sdparm_assoc_arr[assoc]);
-        printf("    desig_type: %s,  code_set: %s\n",
+        printf("    designator type: %s,  code set: %s\n",
                sdparm_desig_type_arr[desig_type], sdparm_code_set_arr[c_set]);
         if (piv && ((1 == assoc) || (2 == assoc)))
-            printf("     transport: %s\n", sdparm_transport_proto_arr[p_id]);
+            printf("     transport: %s\n",
+                   sg_get_trans_proto_str(p_id, sizeof(b), b));
         /* printf("    associated with the %s\n", sdparm_assoc_arr[assoc]); */
         switch (desig_type) {
         case 0: /* vendor specific */
@@ -221,7 +262,7 @@ static int decode_dev_ids(const char * print_if_found, unsigned char * buff,
             if (! long_out) {
                 if ((8 != i_len) && (12 != i_len) && (16 != i_len)) {
                     fprintf(stderr, "      << expect 8, 12 and 16 byte "
-                            "ids, got %d>>\n", i_len);
+                            "EUI, got %d>>\n", i_len);
                     dStrHex((const char *)ip, i_len, 0);
                     break;
                 }
@@ -246,7 +287,7 @@ static int decode_dev_ids(const char * print_if_found, unsigned char * buff,
                         id_ext <<= 8;
                     id_ext |= ip[m];
                 }
-                printf("      Identifier extension: 0x%llx\n", id_ext);
+                printf("      Identifier extension: 0x%" PRIx64 "\n", id_ext);
             } else if ((8 != i_len) && (12 != i_len)) {
                 fprintf(stderr, "      << can only decode 8, 12 and 16 "
                         "byte ids>>\n");
@@ -262,29 +303,30 @@ static int decode_dev_ids(const char * print_if_found, unsigned char * buff,
                     vsei <<= 8;
                 vsei |= ip[ci_off + 3 + m];
             }
-            printf("      Vendor Specific Extension Identifier: 0x%llx\n",
-                   vsei);
+            printf("      Vendor Specific Extension Identifier: 0x%" PRIx64
+                   "\n", vsei);
             if (12 == i_len) {
-                d_id = ((ip[8] << 24) | (ip[9] << 16) | (ip[10] << 8) |
-                        ip[11]);
+                d_id = (((unsigned int)ip[8] << 24) | (ip[9] << 16) |
+                        (ip[10] << 8) | ip[11]);
                 printf("      Directory ID: 0x%x\n", d_id);
             }
             break;
         case 3: /* NAA */
             if (1 != c_set) {
-                fprintf(stderr, "      << expected binary code_set (1)>>\n");
+                fprintf(stderr, "      << unexpected code set %d for "
+                        "NAA>>\n", c_set);
                 dStrHex((const char *)ip, i_len, 0);
                 break;
             }
             naa = (ip[0] >> 4) & 0xff;
             if (! ((2 == naa) || (5 == naa) || (6 == naa))) {
-                fprintf(stderr, "      << expected naa [0x%x]>>\n", naa);
+                fprintf(stderr, "      << unexpected NAA [0x%x]>>\n", naa);
                 dStrHex((const char *)ip, i_len, 0);
                 break;
             }
             if (2 == naa) {
                 if (8 != i_len) {
-                    fprintf(stderr, "      << expected NAA 2 identifier "
+                    fprintf(stderr, "      << unexpected NAA 2 identifier "
                             "length: 0x%x>>\n", i_len);
                     dStrHex((const char *)ip, i_len, 0);
                     break;
@@ -308,12 +350,12 @@ static int decode_dev_ids(const char * print_if_found, unsigned char * buff,
                 printf("\n");
             } else if (5 == naa) {
                 if (8 != i_len) {
-                    fprintf(stderr, "      << expected NAA 5 identifier "
+                    fprintf(stderr, "      << unexpected NAA 5 identifier "
                             "length: 0x%x>>\n", i_len);
                     dStrHex((const char *)ip, i_len, 0);
                     break;
                 }
-                c_id = (((ip[0] & 0xf) << 20) | (ip[1] << 12) | 
+                c_id = (((ip[0] & 0xf) << 20) | (ip[1] << 12) |
                         (ip[2] << 4) | ((ip[3] & 0xf0) >> 4));
                 vsei = ip[3] & 0xf;
                 for (m = 1; m < 5; ++m) {
@@ -322,8 +364,8 @@ static int decode_dev_ids(const char * print_if_found, unsigned char * buff,
                 }
                 if (long_out) {
                     printf("      NAA 5, IEEE Company_id: 0x%x\n", c_id);
-                    printf("      Vendor Specific Identifier: 0x%llx\n",
-                           vsei);
+                    printf("      Vendor Specific Identifier: 0x%" PRIx64
+                           "\n", vsei);
                     printf("      [0x");
                     for (m = 0; m < 8; ++m)
                         printf("%02x", (unsigned int)ip[m]);
@@ -336,12 +378,12 @@ static int decode_dev_ids(const char * print_if_found, unsigned char * buff,
                 }
             } else if (6 == naa) {
                 if (16 != i_len) {
-                    fprintf(stderr, "      << expected NAA 6 identifier "
+                    fprintf(stderr, "      << unexpected NAA 6 identifier "
                             "length: 0x%x>>\n", i_len);
                     dStrHex((const char *)ip, i_len, 0);
                     break;
                 }
-                c_id = (((ip[0] & 0xf) << 20) | (ip[1] << 12) | 
+                c_id = (((ip[0] & 0xf) << 20) | (ip[1] << 12) |
                         (ip[2] << 4) | ((ip[3] & 0xf0) >> 4));
                 vsei = ip[3] & 0xf;
                 for (m = 1; m < 5; ++m) {
@@ -350,8 +392,8 @@ static int decode_dev_ids(const char * print_if_found, unsigned char * buff,
                 }
                 if (long_out) {
                     printf("      NAA 6, IEEE Company_id: 0x%x\n", c_id);
-                    printf("      Vendor Specific Identifier: 0x%llx\n",
-                           vsei);
+                    printf("      Vendor Specific Identifier: 0x%" PRIx64
+                           "\n", vsei);
                     vsei = 0;
                     for (m = 0; m < 8; ++m) {
                         if (m > 0)
@@ -359,7 +401,7 @@ static int decode_dev_ids(const char * print_if_found, unsigned char * buff,
                         vsei |= ip[8 + m];
                     }
                     printf("      Vendor Specific Identifier Extension: "
-                           "0x%llx\n", vsei);
+                           "0x%" PRIx64 "\n", vsei);
                     printf("      [0x");
                     for (m = 0; m < 16; ++m)
                         printf("%02x", (unsigned int)ip[m]);
@@ -382,7 +424,7 @@ static int decode_dev_ids(const char * print_if_found, unsigned char * buff,
             d_id = ((ip[2] << 8) | ip[3]);
             printf("      Relative target port: 0x%x\n", d_id);
             break;
-        case 5: /* Target port group */
+        case 5: /* (primary) Target port group */
             if ((1 != c_set) || (1 != assoc) || (4 != i_len)) {
                 fprintf(stderr, "      << expected binary code_set, target "
                         "port association, length 4>>\n");
@@ -438,7 +480,8 @@ static int decode_dev_ids(const char * print_if_found, unsigned char * buff,
     return 0;
 }
 
-static int decode_mode_policy_vpd(unsigned char * buff, int len)
+static int
+decode_mode_policy_vpd(unsigned char * buff, int len)
 {
     int k, bump;
     unsigned char * ucp;
@@ -468,7 +511,8 @@ static int decode_mode_policy_vpd(unsigned char * buff, int len)
     return 0;
 }
 
-static int decode_man_net_vpd(unsigned char * buff, int len)
+static int
+decode_man_net_vpd(unsigned char * buff, int len)
 {
     int k, bump, na_len;
     unsigned char * ucp;
@@ -498,8 +542,87 @@ static int decode_man_net_vpd(unsigned char * buff, int len)
     return 0;
 }
 
-static int decode_scsi_ports_vpd(unsigned char * buff, int len,
-                                 int long_out, int quiet)
+static int
+decode_proto_lu_vpd(unsigned char * buff, int len)
+{
+    int k, bump, rel_port, desc_len, proto;
+    unsigned char * ucp;
+
+    if (len < 4) {
+        fprintf(stderr, "Protocol-specific logical unit information VPD "
+                "page length too short=%d\n", len);
+        return SG_LIB_CAT_MALFORMED;
+    }
+    len -= 4;
+    ucp = buff + 4;
+    for (k = 0; k < len; k += bump, ucp += bump) {
+        rel_port = (ucp[0] << 8) + ucp[1];
+        printf("Relative port=%d\n", rel_port);
+        proto = ucp[2] & 0xf;
+        desc_len = (ucp[6] << 8) + ucp[7];
+        bump = 8 + desc_len;
+        if ((k + bump) > len) {
+            fprintf(stderr, "Protocol-specific logical unit information VPD "
+                    "page, short descriptor length=%d, left=%d\n", bump,
+                    (len - k));
+            return SG_LIB_CAT_MALFORMED;
+        }
+        if (desc_len > 0) {
+            switch (proto) {
+            case TPROTO_SAS:
+                printf(" Protocol identifier: SAS\n");
+                printf(" TLR control supported: %d\n", !!(ucp[8] & 0x1));
+                break;
+            default:
+                fprintf(stderr, "Unexpected proto=%d\n", proto);
+                dStrHex((const char *)ucp, bump, 1);
+                break;
+            }
+        }
+    }
+    return 0;
+}
+
+static int
+decode_proto_port_vpd(unsigned char * buff, int len)
+{
+    int k, bump, rel_port, desc_len, proto;
+    unsigned char * ucp;
+
+    if (len < 4) {
+        fprintf(stderr, "Protocol-specific port information VPD "
+                "page length too short=%d\n", len);
+        return SG_LIB_CAT_MALFORMED;
+    }
+    len -= 4;
+    ucp = buff + 4;
+    for (k = 0; k < len; k += bump, ucp += bump) {
+        rel_port = (ucp[0] << 8) + ucp[1];
+        printf("Relative port=%d\n", rel_port);
+        proto = ucp[2] & 0xf;
+        desc_len = (ucp[6] << 8) + ucp[7];
+        bump = 8 + desc_len;
+        if ((k + bump) > len) {
+            fprintf(stderr, "Protocol-specific port information VPD "
+                    "page, short descriptor length=%d, left=%d\n", bump,
+                    (len - k));
+            return SG_LIB_CAT_MALFORMED;
+        }
+        if (desc_len > 0) {
+            switch (proto) {
+            case TPROTO_SAS:
+            default:
+                fprintf(stderr, "Unexpected proto=%d\n", proto);
+                dStrHex((const char *)ucp, bump, 1);
+                break;
+            }
+        }
+    }
+    return 0;
+}
+
+static int
+decode_scsi_ports_vpd(unsigned char * buff, int len, int long_out, int quiet)
 {
     int k, bump, rel_port, ip_tid_len, tpd_len, res;
     unsigned char * ucp;
@@ -521,7 +644,7 @@ static int decode_scsi_ports_vpd(unsigned char * buff, int len,
             return SG_LIB_CAT_MALFORMED;
         }
         if (ip_tid_len > 0) {
-            /* 
+            /*
              * SCSI devices that are both target and initiator are rare.
              * Only target devices can receive this command, so if they
              * are also initiators then print out the "Initiator port
@@ -548,21 +671,46 @@ static int decode_scsi_ports_vpd(unsigned char * buff, int len,
     return 0;
 }
 
-static int decode_ext_inq_vpd(unsigned char * buff, int len)
+static int
+decode_ext_inq_vpd(unsigned char * buff, int len, int quiet)
 {
     if (len < 7) {
         fprintf(stderr, "Extended INQUIRY data VPD page length too "
                 "short=%d\n", len);
         return SG_LIB_CAT_MALFORMED;
     }
-    printf("  SPT: %d  GRD_CHK: %d  APP_CHK: %d  REF_CHK: %d\n",
-           ((buff[4] >> 3) & 0x7), !!(buff[4] & 0x4), !!(buff[4] & 0x2),
-           !!(buff[4] & 0x1));
-    printf("  GRP_SUP: %d  PRIOR_SUP: %d  HEADSUP: %d  ORDSUP: %d  "
-           "SIMPSUP: %d\n", !!(buff[5] & 0x10), !!(buff[5] & 0x8),
-           !!(buff[5] & 0x4), !!(buff[5] & 0x2), !!(buff[5] & 0x1));
-    printf("  CORR_D_SUP=%d NV_SUP=%d V_SUP=%d\n", !!(buff[6] & 0x80),
-           !!(buff[6] & 0x2), !!(buff[6] & 0x1));
+    if (quiet) {
+        printf("spt=%d\n", ((buff[4] >> 3) & 0x7));
+        printf("grd_chk=%d\n", !!(buff[4] & 0x4));
+        printf("app_chk=%d\n", !!(buff[4] & 0x2));
+        printf("ref_chk=%d\n", !!(buff[4] & 0x1));
+        printf("uask_sup=%d\n", !!(buff[5] & 0x20));
+        printf("group_sup=%d\n", !!(buff[5] & 0x10));
+        printf("prior_sup=%d\n", !!(buff[5] & 0x8));
+        printf("headsup=%d\n", !!(buff[5] & 0x4));
+        printf("ordsup=%d\n", !!(buff[5] & 0x2));
+        printf("simpsup=%d\n", !!(buff[5] & 0x1));
+        printf("wu_sup=%d\n", !!(buff[6] & 0x8));
+        printf("crd_sup=%d\n", !!(buff[6] & 0x4));
+        printf("nv_sup=%d\n", !!(buff[6] & 0x2));
+        printf("v_sup=%d\n", !!(buff[6] & 0x1));
+        printf("luiclr=%d\n", !!(buff[7] & 0x1));
+        printf("cbcs=%d\n", !!(buff[8] & 0x1));
+        printf("mitnd=%d\n", (buff[9] & 0xf));
+    } else {
+        printf("  SPT=%d GRD_CHK=%d APP_CHK=%d REF_CHK=%d\n",
+               ((buff[4] >> 3) & 0x7), !!(buff[4] & 0x4), !!(buff[4] & 0x2),
+               !!(buff[4] & 0x1));
+        printf("  UASK_SUP=%d GROUP_SUP=%d PRIOR_SUP=%d HEADSUP=%d ORDSUP=%d "
+               "SIMPSUP=%d\n", !!(buff[5] & 0x20), !!(buff[5] & 0x10),
+               !!(buff[5] & 0x8), !!(buff[5] & 0x4), !!(buff[5] & 0x2),
+               !!(buff[5] & 0x1));
+        printf("  WU_SUP=%d CRD_SUP=%d NV_SUP=%d V_SUP=%d LUICLR=%d "
+               "CBCS=%d\n", !!(buff[6] & 0x8), !!(buff[6] & 0x4),
+               !!(buff[6] & 0x2), !!(buff[6] & 0x1), !!(buff[7] & 0x1),
+               !!(buff[8] & 0x1));
+        printf("  Multi I_T nexus microcode download=%d\n", buff[9] & 0xf);
+    }
     return 0;
 }
 
@@ -625,7 +773,8 @@ static int decode_ata_info_vpd(unsigned char * buff, int len, int long_out,
     return 0;
 }
 
-static int decode_block_limits_vpd(unsigned char * buff, int len)
+static int
+decode_block_limits_vpd(unsigned char * buff, int len)
 {
     unsigned int u;
 
@@ -636,16 +785,45 @@ static int decode_block_limits_vpd(unsigned char * buff, int len)
     }
     u = (buff[6] << 8) | buff[7];
     printf("  Optimal transfer length granularity: %u blocks\n", u);
-    u = (buff[8] << 24) | (buff[9] << 16) | (buff[10] << 8) |
-        buff[11];
+    u = ((unsigned int)buff[8] << 24) | (buff[9] << 16) |
+        (buff[10] << 8) | buff[11];
     printf("  Maximum transfer length: %u blocks\n", u);
-    u = (buff[12] << 24) | (buff[13] << 16) | (buff[14] << 8) |
-        buff[15];
+    u = ((unsigned int)buff[12] << 24) | (buff[13] << 16) |
+        (buff[14] << 8) | buff[15];
     printf("  Optimal transfer length: %u blocks\n", u);
+    if (len > 19) {     /* added in sbc3r09 */
+        u = ((unsigned int)buff[16] << 24) | (buff[17] << 16) |
+            (buff[18] << 8) | buff[19];
+        printf("  Maximum prefetch, xdread, xdwrite transfer length: %u "
+               "blocks\n", u);
+    }
     return 0;
 }
 
-static int decode_tape_dev_caps_vpd(unsigned char * buff, int len)
+static int
+decode_block_dev_chars_vpd(unsigned char * buff, int len)
+{
+    unsigned int u;
+
+    if (len < 64) {
+        fprintf(stderr, "Block device capabilities VPD page length too "
+                "short=%d\n", len);
+        return SG_LIB_CAT_MALFORMED;
+    }
+    u = (buff[4] << 8) | buff[5];
+    if (0 == u)
+        printf("  Medium rotation rate is not reported\n");
+    else if (1 == u)
+        printf("  Non-rotating medium (e.g. solid state)\n");
+    else if ((u < 0x401) || (0xffff == u))
+        printf("  Reserved [0x%x]\n", u);
+    else
+        printf("  Nominal rotation rate: %d rpm\n", u);
+    return 0;
+}
+
+static int
+decode_tape_dev_caps_vpd(unsigned char * buff, int len)
 {
     if (len < 6) {
         fprintf(stderr, "Sequential access device capabilities VPD page "
@@ -656,7 +834,21 @@ static int decode_tape_dev_caps_vpd(unsigned char * buff, int len)
     return 0;
 }
 
-static int decode_tapealert_supported_vpd(unsigned char * b, int len)
+static int
+decode_tape_man_ass_sn_vpd(unsigned char * buff, int len)
+{
+    if (len < 64) {
+        fprintf(stderr, "Manufacturer-assigned serial number VPD page "
+                "length too short=%d\n", len);
+        return SG_LIB_CAT_MALFORMED;
+    }
+    printf("  Manufacturer-assigned serial number: %.*s\n",
+                   len - 4, buff + 4);
+    return 0;
+}
+
+static int
+decode_tapealert_supported_vpd(unsigned char * b, int len)
 {
     if (len < 12) {
         fprintf(stderr, "TapeAlert supported flags length too short=%d\n",
@@ -699,17 +891,17 @@ static int decode_tapealert_supported_vpd(unsigned char * b, int len)
 }
 
 /* Returns 0 if successful, else error */
-int sdp_process_vpd_page(int sg_fd, int pn, int spn,
-                         const struct sdparm_opt_coll * opts,
-                         int req_pdt, int verbose)
+int
+sdp_process_vpd_page(int sg_fd, int pn, int spn,
+                     const struct sdparm_opt_coll * opts, int req_pdt)
 {
     int res, len, k, verb, dev_pdt, pdt;
     unsigned char b[VPD_ATA_INFO_RESP_LEN];
     int sz;
     unsigned char * up;
-    const struct sdparm_values_name_t * vnp;
+    const struct sdparm_vpd_page_t * vpp;
 
-    verb = (verbose > 0) ? verbose - 1 : 0;
+    verb = (opts->verbose > 0) ? opts->verbose - 1 : 0;
     sz = sizeof(b);
     memset(b, 0, sz);
     if (pn < 0) {
@@ -744,13 +936,13 @@ int sdp_process_vpd_page(int sg_fd, int pn, int spn,
         }
         if (len > 0) {
             for (k = 0; k < len; ++k) {
-                vnp = sdp_get_vpd_detail(b[4 + k], -1, pdt);
-                if (vnp) {
+                vpp = sdp_get_vpd_detail(b[4 + k], -1, pdt);
+                if (vpp) {
                     if (opts->long_out)
                         printf("  [0x%02x] %s [%s]\n", b[4 + k],
-                               vnp->name, vnp->acron);
+                               vpp->name, vpp->acron);
                     else
-                        printf("  %s [%s]\n", vnp->name, vnp->acron);
+                        printf("  %s [%s]\n", vpp->name, vpp->acron);
                 } else
                     printf("  0x%x\n", b[4 + k]);
             }
@@ -783,38 +975,6 @@ int sdp_process_vpd_page(int sg_fd, int pn, int spn,
         if (res)
             return res;
         break;
-    case VPD_BLOCK_LIMITS:      /* same value as VPD_SA_DEV_CAP */
-        {
-            int tape;
-            const char * vpd_name;
-
-            tape = ((1 == pdt) || (8 == pdt)) ? 1 : 0;
-            vpd_name = tape ? "Sequential access device capabilities" :
-                              "Block limits";
-            if (b[1] != pn)
-                goto dumb_inq;
-            len = (b[2] << 8) + b[3];
-            if (len > sz) {
-                fprintf(stderr, "Response to %s VPD page truncated\n",
-                        vpd_name);
-                len = sz;
-            }
-            if (opts->long_out)
-                printf("%s [0xb0] VPD page:\n", vpd_name);
-            else
-                printf("%s VPD page:\n", vpd_name);
-            if (opts->hex) {
-                dStrHex((const char *)b, len + 4, 0);
-                return 0;
-            }
-            if (tape)
-                res = decode_tape_dev_caps_vpd(b, len + 4);
-            else
-                res = decode_block_limits_vpd(b, len + 4);
-            if (res)
-                return res;
-        }
-        break;
     case VPD_DEVICE_ID:
         if (b[1] != pn)
             goto dumb_inq;
@@ -862,13 +1022,13 @@ int sdp_process_vpd_page(int sg_fd, int pn, int spn,
         }
         if (opts->long_out)
             printf("Extended inquiry data [0x86] VPD page:\n");
-        else
+        else if (! opts->quiet)
             printf("Extended inquiry data VPD page:\n");
         if (opts->hex) {
             dStrHex((const char *)b, len + 4, 0);
             return 0;
         }
-        res = decode_ext_inq_vpd(b, len + 4);
+        res = decode_ext_inq_vpd(b, len + 4, opts->quiet);
         if (res)
             return res;
         break;
@@ -914,6 +1074,50 @@ int sdp_process_vpd_page(int sg_fd, int pn, int spn,
         if (res)
             return res;
         break;
+    case VPD_PROTO_LU:
+        if (b[1] != pn)
+            goto dumb_inq;
+        len = (b[2] << 8) + b[3];
+        if (len > sz) {
+            fprintf(stderr, "Response to Protocol-specific logical unit "
+                    "information VPD page truncated\n");
+            len = sz;
+        }
+        if (opts->long_out)
+            printf("Protocol-specific logical unit information [0x90] VPD "
+                   "page:\n");
+        else
+            printf("Protocol-specific logical unit information VPD page:\n");
+        if (opts->hex) {
+            dStrHex((const char *)b, len + 4, 0);
+            return 0;
+        }
+        res = decode_proto_lu_vpd(b, len + 4);
+        if (res)
+            return res;
+        break;
+    case VPD_PROTO_PORT:
+        if (b[1] != pn)
+            goto dumb_inq;
+        len = (b[2] << 8) + b[3];
+        if (len > sz) {
+            fprintf(stderr, "Response to Protocol-specific port "
+                    "information VPD page truncated\n");
+            len = sz;
+        }
+        if (opts->long_out)
+            printf("Protocol-specific port information [0x91] VPD "
+                   "page:\n");
+        else
+            printf("Protocol-specific port information VPD page:\n");
+        if (opts->hex) {
+            dStrHex((const char *)b, len + 4, 0);
+            return 0;
+        }
+        res = decode_proto_port_vpd(b, len + 4);
+        if (res)
+            return res;
+        break;
     case VPD_SCSI_PORTS:
         if (b[1] != pn)
             goto dumb_inq;
@@ -960,22 +1164,6 @@ int sdp_process_vpd_page(int sg_fd, int pn, int spn,
             printf("\n");
         }
         break;
-    case VPD_TA_SUPPORTED:
-        if (b[1] != pn)
-            goto dumb_inq;
-        len = b[3];
-        if (opts->long_out)
-            printf("TapeAlert supported flags [0xb2] VPD page:\n");
-        else
-            printf("TapeAlert supported flags VPD page:\n");
-        if (opts->hex) {
-            dStrHex((const char *)b, len + 4, 0);
-            return 0;
-        }
-        res = decode_tapealert_supported_vpd(b, len + 4);
-        if (res)
-            return res;
-        break;
     case VPD_UNIT_SERIAL_NUM:
         if (b[1] != pn)
             goto dumb_inq;
@@ -995,17 +1183,140 @@ int sdp_process_vpd_page(int sg_fd, int pn, int spn,
         } else
             printf("  <empty>\n");
         break;
+    case 0xb0:          /* VPD page depends on pdt */
+        {
+            int osd = 0;
+            int sbc = 0;
+            int ssc = 0;
+            const char * vpd_name;
+
+            switch (pdt)
+            {
+            case PDT_DISK: case PDT_WO: case PDT_OPTICAL:
+                vpd_name = "Block limits";
+                sbc = 1;
+                break;
+            case PDT_TAPE: case PDT_MCHANGER:
+                vpd_name = "Sequential access device capabilities";
+                ssc = 1;
+                break;
+            case PDT_OSD:
+                vpd_name = "OSD information";
+                osd = 1;
+                break;
+            default:
+                vpd_name = "unexpected pdt for B0h";
+                break;
+            }
+            if (b[1] != pn)
+                goto dumb_inq;
+            len = (b[2] << 8) + b[3];
+            if (len > sz) {
+                fprintf(stderr, "Response to %s VPD page truncated\n",
+                        vpd_name);
+                len = sz;
+            }
+            if (opts->long_out)
+                printf("%s [0xb0] VPD page:\n", vpd_name);
+            else
+                printf("%s VPD page:\n", vpd_name);
+            if (opts->hex) {
+                dStrHex((const char *)b, len + 4, 0);
+                return 0;
+            }
+            if (ssc)
+                res = decode_tape_dev_caps_vpd(b, len + 4);
+            else if (sbc)
+                res = decode_block_limits_vpd(b, len + 4);
+            else
+                dStrHex((const char *)b, len + 4, 0);
+            if (res)
+                return res;
+        }
+        break;
+    case 0xb1:          /* VPD page depends on pdt */
+        {
+            int adc = 0;
+            int osd = 0;
+            int sbc = 0;
+            int ssc = 0;
+            const char * vpd_name;
+
+            switch (pdt)
+            {
+            case PDT_DISK: case PDT_WO: case PDT_OPTICAL:
+                vpd_name = "Block device characteristics";
+                sbc = 1;
+                break;
+            case PDT_TAPE: case PDT_MCHANGER:
+                vpd_name = "Manufactured assigned serial number";
+                ssc = 1;
+                break;
+            case PDT_OSD:
+                vpd_name = "Security token";
+                osd = 1;
+                break;
+            case PDT_ADC:
+                vpd_name = "Manufactured assigned serial number";
+                adc = 1;
+                break;
+            default:
+                vpd_name = "unexpected pdt for B1h";
+                break;
+            }
+            if (b[1] != pn)
+                goto dumb_inq;
+            len = (b[2] << 8) + b[3];
+            if (len > sz) {
+                fprintf(stderr, "Response to %s VPD page truncated\n",
+                        vpd_name);
+                len = sz;
+            }
+            if (opts->long_out)
+                printf("%s [0xb1] VPD page:\n", vpd_name);
+            else
+                printf("%s VPD page:\n", vpd_name);
+            if (opts->hex) {
+                dStrHex((const char *)b, len + 4, 0);
+                return 0;
+            }
+            if (ssc || adc)
+                res = decode_tape_man_ass_sn_vpd(b, len + 4);
+            else if (sbc)
+                res = decode_block_dev_chars_vpd(b, len + 4);
+            else
+                dStrHex((const char *)b, len + 4, 0);
+            if (res)
+                return res;
+        }
+        break;
+    case 0xb2:          /* VPD page depends on pdt, only VPD_TA_SUPPORTED */
+        if (b[1] != pn)
+            goto dumb_inq;
+        len = b[3];
+        if (opts->long_out)
+            printf("TapeAlert supported flags [0xb2] VPD page:\n");
+        else
+            printf("TapeAlert supported flags VPD page:\n");
+        if (opts->hex) {
+            dStrHex((const char *)b, len + 4, 0);
+            return 0;
+        }
+        res = decode_tapealert_supported_vpd(b, len + 4);
+        if (res)
+            return res;
+        break;
     default:
         if (b[1] != pn)
             goto dumb_inq;
         len = (b[2] << 8) + b[3] + 4;
-        vnp = sdp_get_vpd_detail(pn, -1, pdt);
-        if (vnp)
-            fprintf(stderr, "%s VPD page in hex:\n", vnp->name);
+        vpp = sdp_get_vpd_detail(pn, -1, pdt);
+        if (vpp)
+            fprintf(stderr, "%s VPD page in hex:\n", vpp->name);
         else
             fprintf(stderr, "VPD page 0x%x in hex:\n", pn);
         if (len > (int)sizeof(b)) {
-            if (verbose)
+            if (opts->verbose)
                 fprintf(stderr, "page length=%d too long, trim\n", len);
             len = sizeof(b);
         }
